/*
 ftrig.c
  integer fraction routines

  Created: 2008.07.15
  Revised: 2014.03.09

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

2014.02.27
  usqrt/usqrt64 optimizations, fixed & optimized vfix16mul.
  also optimized vfix16div in Intel asm.
2014.02.28
  hypot funcs in assembly (Intel).
2014.03.01
  fastest usqrt/usqrt64 ever.
*/

#include "../enviro64.h"
#include "ftrig.h"

#define NO_FPU		/* integer usqrt funcs actually faster now */

#ifdef env68k
 #pragma segment Lib
#else
#if defined(__POCC__) && envx86
__declspec(naked)
#endif
ifixed vfix16mul(ifixed a,
	ifixed b)
{
#if defined(__POCC__) && envx86
	__asm
	{
		push	ebx
		mov	eax,[esp+8]
		imul	dword [esp+12]
		mov	ebx,edx
		sar	ebx,31
		lea	ecx,[ebx+0x8000]
		add	eax,ecx
		adc	edx,0
		shrd	eax,edx,16
		sar	edx,16
		mov	ecx,edx
		cdq
		cmp	ecx,edx
		je	eo

		mov	eax,ebx
		xor	eax,0x7FFFFFFF
	eo:	pop	ebx
		ret
	}
#else
	int64		v;

	(void)muls64s32(v,a,b);
	/* prev. version incorrectly added -0x8000 in negative case */
	b = (ifixed)(int64hi(v)>>31);
	(void)int64addu32(v,b+0x8000);
	int64shr(v,16);
	if (((a=int64lo(v))>>31)!=int64hi(v))
		return b^0x7FFFFFFF; /* overflow */
	return a;
#endif
}

#if defined(__POCC__) && envx86
__declspec(naked)
#endif
ifixed vfix16div(ifixed a,
	ifixed b)
{
#if defined(__POCC__) && envx86
	__asm
	{ /* this'd be shorter if overflow flagged instead of making an exception */
		push	ebx
		mov	eax,[esp+12]
		cdq
		mov	ebx,edx
		xor	eax,edx
		sub	eax,edx
		mov	ecx,eax
		mov	eax,[esp+8]
		cdq
		xor	ebx,edx
		xor	eax,edx
		sub	eax,edx
		mov	edx,eax
		shr	edx,15
		shl	eax,17
		add	eax,ecx
		adc	edx,0
		cmp	edx,ecx
		jb	nv

		mov	eax,ebx
		xor	eax,0x7FFFFFFF
		jmp	eo

	nv:	shr	edx,1
		rcr	eax,1
		div	ecx
		xor	eax,ebx
		sub	eax,ebx
	eo:	pop	ebx
		ret
	}
#else
	int64		v;
	int32		r = a>>31;

	if (!b)
		return r^0x7FFFFFFF;
	r^= (ABS(b)>>1)+r;
	(void)int64qset(v,a);
	(void)int64adds32(v,r);
	(void)int64divs32(v,b);

	return int64toi(v);
#endif
}

#if !defined(NO_FPU) && defined(__POCC__) && envx86
/* faster to use the FPU */
static float	nint64min = 9223372036854775808.;

uint32 __declspec(naked) usqrt64(uint64 n)
{
	__asm
	{ /* The trick to loading unsigned integers: sign-invert & compensate.
	 (The other trick, but probably slower: load subnormally.)
	 One problem: fistp rounds fractions of exactly 0.5 toward *zero*.
	 Rounding will have to be done manually.

	 Observation, setting the P4 prec mode to 'reserved': same as 64-bit.
	*/
		xor	byte [esp+11],128
		fild	qword [esp+4]
		fadd	nint64min
		fsqrt
		push	ecx
		push	ecx
		fistp	qword [esp]
		pop	eax
		pop	ecx
		sub	eax,ecx		;largest case: ecx=1:eax=0
		or	ecx,eax
		inc	eax
		jz	eo

		xor	byte [esp+11],128
		mul	ecx
		sub	eax,[esp+4]
		sbb	edx,[esp+8]
		mov	eax,0
	eo:	adc	eax,ecx
		ret
	}
}

uint32 __declspec(naked) usqrt(uint32 n)
{
	__asm
	{
		push	0
		push	dword [esp+8]
		fild	qword [esp]
		pop	ecx
		fsqrt
		fistp	dword [esp]
		pop	ecx
		lea	eax,[ecx+1]
		mul	ecx
		sub	eax,[esp+4]
		sbb	edx,0
		mov	eax,0
		adc	eax,ecx
		ret
	}
}
#else
/* square root half (16 bit) table */
static const uint16	srhlut[256] =
{
	1,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,
	32900,33152,33404,33656,33904,34152,34396,34640,
	34880,35120,35356,35596,35828,36064,36296,36524,
	36756,36980,37208,37432,37656,37880,38100,38320,
	38536,38752,38968,39184,39400,39612,39820,40032,
	40240,40448,40656,40860,41068,41272,41472,41676,
	41876,42076,42276,42472,42668,42864,43060,43256,
	43448,43640,43832,44024,44216,44404,44592,44780,
	44968,45152,45340,45524,45708,45892,46072,46256,
	46436,46616,46796,46976,47152,47332,47508,47684,
	47860,48032,48208,48380,48556,48728,48900,49072,
	49240,49412,49580,49748,49920,50084,50252,50420,
	50584,50752,50916,51080,51244,51408,51572,51732,
	51896,52056,52216,52380,52540,52696,52856,53016,
	53172,53332,53488,53644,53800,53956,54112,54268,
	54420,54576,54728,54880,55032,55188,55336,55488,
	55640,55792,55940,56092,56240,56388,56536,56684,
	56832,56980,57128,57276,57420,57568,57712,57856,
	58004,58148,58292,58436,58580,58720,58864,59008,
	59148,59288,59432,59572,59712,59852,59992,60132,
	60272,60412,60552,60688,60828,60964,61100,61240,
	61376,61512,61648,61784,61920,62056,62192,62324,
	62460,62592,62728,62860,62996,63128,63260,63392,
	63524,63656,63788,63920,64052,64180,64312,64444,
	64572,64704,64832,64960,65092,65220,65348,65535
};

 #if defined(__POCC__) && envx86
__declspec(naked)
 #endif
uint32 usqrt64(uint64 n)
{
 #if defined(__POCC__) && envx86
	__asm
	{
		mov	edx,[esp+8]
		mov	cl,-1	/* default absolutely needed for zero */
		bsr	ecx,edx
		jz	usqrt

		not	cl
		add	cl,32
		and	ecx,0x3E
		mov	eax,[esp+4]
		shld	edx,eax,cl
		shr	cl,1
		not	cl
		add	cl,16
		mov	eax,edx
		shr	edx,24
		push	cx
		mov	cx,[srhlut+edx+edx]
		mov	dl,0
		div	ecx
		lea	eax,[ecx+eax+1]
		pop	cx
		shl	eax,cl
		sbb	ecx,ecx
		or	ecx,eax
		mov	eax,[esp+4]
		mov	edx,[esp+8]
		cmp	ecx,edx
		ja	nv

		or	eax,-1
		ret

	nv:	div	ecx
		dec	ecx
		add	eax,ecx
		rcr	eax,1
		lea	ecx,[eax+1]
		mul	ecx
		sub	eax,[esp+4]
		sbb	edx,[esp+8]
		adc	ecx,-1
		mov	eax,ecx
		ret
	}
 #else
	uint64		t;
	/* get the highest ordered bit, and shift up... */
	uint32		r, s, a;

	if ((a=int64hi(n))==0)
		return usqrt(int64lo(n));
	s = cntlz32(a)&~1;
	a = a<<s|int64lo(n)>>(32-s); /* LONG_LONG alt. is worse on 32-bit */
	s = 16+~(s>>1); /* reset scale when finished */
	r = srhlut[a>>24];
	/* divide and average - -1/+1 to prevent overflow & pre-rounding */
	r = (a/r+r-1)<<s;
	t = n; (void)uint64divu32(t,r+1); (void)uint64addu32(t,r); uint64half(t);
	if ((r=int64lo(t))==0xFFFFFFFFu)
		return r;
	(void)mulu64u32(t,r,r+1);
	return r+((uint32)uint64sub(t,n)>>31);
 #endif
}

 #if defined(__POCC__) && envx86
__declspec(naked)
 #endif
uint32 usqrt(uint32 n)
{
 #if defined(__POCC__) && envx86
	__asm
	{ /* no branches! */
		mov	eax,[esp+4]
		mov	cl,-1	/* default absolutely needed for zero */
		bsr	ecx,eax
		not	cl
		add	cl,32
		and	ecx,0x3E
		shl	eax,cl
		shr	cl,1
		inc	cl
		mov	edx,eax
		shr	edx,24
		push	cx
		mov	cx,[srhlut+edx+edx]
		mov	dl,0
		div	ecx	/* 16-bit version isn't faster! */
		lea	eax,[ecx+eax-1]
		pop	cx
		shr	eax,cl
		lea	ecx,[eax+1]	;mov/inc is slower
		mul	ecx
		sub	eax,[esp+4]
		adc	ecx,-1
		mov	eax,ecx
		ret
	}
 #else
 /* to do this effectively, we get the highest ordered bit, and shift up... */
	uint32		a, r, s = cntlz32(n)&~1;

	a = n<<s, r = srhlut[a>>24];
	s = (s>>1)+1; /* reset scale when finished */
	/* the table is large enough that we only have to divide and average once */
	/* make the denom slightly larger to prevent dbz & pre-rounding */
	r = (r+a/(r+1))>>s;
	r+= (signed)((r+1)*r-n)<0;
	return r;
 #endif
}
#endif

#if defined(__POCC__) && envx86
__declspec(naked)
#endif
uint32 ihypot2(int32 i,
	int32 j)
{
#if defined(__POCC__) && envx86
	__asm
	{
		mov	eax,[esp+4]
		imul	eax
		mov	[esp+4],eax
		mov	eax,[esp+8]
		mov	[esp+8],edx
		imul	eax
		add	[esp+4],eax
		adc	[esp+8],edx
		jmp	usqrt64
		ret
	}
#else
	int64		v, t;

	(void)muls64s32(v,i,i);
	(void)muls64s32(t,j,j);
	(void)int64add(v,t);
	return usqrt64(*(uint64*)&v);
#endif
}

#if defined(__POCC__) && envx86
__declspec(naked)
#endif
uint32 uhypot2(uint32 i,
	uint32 j)
{
#if defined(__POCC__) && envx86
	__asm
	{
		mov	eax,[esp+4]
		mul	eax
		mov	[esp+4],eax
		mov	eax,[esp+8]
		mov	[esp+8],edx
		mul	eax
		add	[esp+4],eax
		adc	[esp+8],edx
		jmp	usqrt64
		ret
	}
#else
	uint64		v, t;

	(void)mulu64u32(v,i,i);
	(void)mulu64u32(t,j,j);
	(void)uint64add(v,t);
	return usqrt64(v);
#endif
}

#if defined(__POCC__) && envx86
__declspec(naked)
#endif
uint32 ihypot3(int32 i,
	int32 j,
	int32 k)
{
#if defined(__POCC__) && envx86
	__asm
	{
		mov	eax,[esp+4]
		imul	eax
		mov	[esp+4],eax
		mov	eax,[esp+8]
		mov	ecx,edx
		imul	eax
		add	[esp+4],eax
		adc	ecx,edx
		mov	eax,[esp+12]
		imul	eax
		add	[esp+4],eax
		adc	ecx,edx
		mov	[esp+8],ecx
		jmp	usqrt64
		ret
	}
#else
	int64		v, t;

	(void)muls64s32(v,i,i);
	(void)muls64s32(t,j,j);
	(void)int64add(v,t);
	(void)muls64s32(t,k,k);
	(void)int64add(v,t);
	return usqrt64(*(uint64*)&v);
#endif
}

#if defined(__POCC__) && envx86
__declspec(naked)
#endif
uint32 uhypot3(uint32 i,
	uint32 j,
	uint32 k)
{
#if defined(__POCC__) && envx86
	__asm
	{
		mov	eax,[esp+4]
		mul	eax
		mov	[esp+4],eax
		mov	eax,[esp+8]
		mov	ecx,edx
		mul	eax
		add	[esp+4],eax
		adc	ecx,edx
		mov	eax,[esp+12]
		mul	eax
		add	[esp+4],eax
		adc	ecx,edx
		mov	[esp+8],ecx
		jmp	usqrt64
		ret
	}
#else
	uint64		v, t;

	(void)mulu64u32(v,i,i);
	(void)mulu64u32(t,j,j);
	(void)uint64add(v,t);
	(void)mulu64u32(t,k,k);
	(void)uint64add(v,t);
	return usqrt64(v);
#endif
}
#endif

/* slopes and scales (originally in mgraf.c) */

#if defined(__POCC__) && envx86
__declspec(naked)
#endif
int64 rslopdiv(int64 *px,
	int32 p0x,
	int32 p0y,
	int32 p1x,
	int32 p1y,
	uint16 ky)
{
#if defined(__POCC__) && envx86
 #define px esp+8
 #define p0 esp+12
 #define p1 esp+20 //reused
 #define ky esp+28
	__asm
	{
		push	ebx
		mov	eax,[p1]
		xor	ebx,ebx
		sub	eax,[p0]
		setl	bl		;sign =-(p0x>p1.x)
		neg	ebx
		xor	eax,ebx 	;invert sign (for abs)
		sub	eax,ebx
		mov	edx,eax		;int64qset
		shr	edx,16
		shl	eax,16
		mov	ecx,[p1+4]
		sub	ecx,[p0+4]
		mov	[p1+4],eax
		lea	eax,[ecx-1]
		add	eax,[p1+4]
		adc	edx,0
		cmp	ecx,edx
		ja	nv

		or	eax,-1
		jmp	rf

	nv:	div	ecx
	rf:	mov	[p1+4],eax
		cmp	dword [px],0
		jz	eo

		mov	eax,[p0]	;int64qset(n,p0x)
		mov	ecx,eax
		sar	ecx,16
		shl	eax,16
		mov	[p1],eax
		mov	eax,[ky-2]	;linefrac
		mov	ax,[p0+4]
		not	ax
		add	eax,1
		sbb	eax,0
		mul	dword [p1+4]
		xor	eax,ebx		;isgn
		sub	eax,ebx
		adc	edx,ebx
		xor	edx,ebx
		add	eax,[p1]
		adc	edx,ecx
		mov	ecx,[px]
		mov	[ecx],eax
		mov	[ecx+4],edx
	eo:	mov	eax,[p1+4]	;int64qset
		mov	edx,eax
		shr	edx,16
		shl	eax,16
		xor	eax,ebx		;isgn
		sub	eax,ebx
		adc	edx,ebx
		xor	edx,ebx
		pop	ebx
		ret
	}
 #undef ky
 #undef p1
 #undef p0
 #undef px
#else
	uint64		n, u;
	ufixed		d, r;
	int32		sign =-(p0x>p1.x);

	if ((d=p1.y-p0y)==0)
		r =-1;
	else
	{ /* round up: add denom-1, matching sign */
		r = (p1.x-p0x+sign)^sign;
		int64qset(n,r);
		(void)uint64addu32(n,d-1); uint64divu32(n,d);
		r = int64tou(n);
	}
	if (px)
	{ /* now, to calculate the initial point */
		int64qset(n,p0x);
		linefrac(d,ky,p0y);
		(void)muls64u32(u,r,d);
		if (sign)
			(void)uint64neg(u);
		(void)int64add(n,u);
		*px = n;
	}
	int64qset(n,r);
	if (sign)
		uint64neg(n);
	return n;
#endif
}

#if defined(__POCC__) && envx86
__declspec(naked)
#endif
uint64 rincdiv(int64 *py,
	ifixed p0y,
	ifixed p1y,
	uint16 ky)
{
#if defined(__POCC__) && envx86
 #define py esp+8
 #define p0y esp+12
 #define p1y esp+16 //reused
 #define ky esp+20
	__asm
	{
		push	ebx
		mov	ecx,[p1y]
		sub	ecx,[p0y]
		mov	eax,0x80000000
		xor	edx,edx
		div	ecx
		mov	[p1y],eax
		lea	eax,[ecx-1]
		div	ecx
		mov	ebx,[py]
		test	ebx,ebx
		jz	eo

		mov	ecx,[ky-2]
		mov	cx,[p0y]
		not	cx
		add	ecx,1
		sbb	ecx,0
		push	eax
		mul	ecx
		mov	[ebx],eax
		mov	eax,ecx
		mov	ecx,edx
		mul	dword [p1y+4]	;+4 because of the saved eax
		add	eax,ecx
		mov	[ebx+4],eax
		pop	eax
	eo:	mov	edx,[p1y]
		shld	edx,eax,16
		shl	eax,16
		pop	ebx
		ret
	}
 #undef ky
 #undef p1y
 #undef p0y
 #undef py
#else
	uint64		n;
	ufixed		d = p1y-p0y;

	int64sethl(n,0x80000000,d-1);
	uint64divu32(n,d);
	if (py)
	{
		linefrac(d,ky,p0y);
		*py = n;
		int64mulu32(*py,d);
	}
	int64shl(n,16);
	return n;
#endif
}

#if defined(__POCC__) && envx86
__declspec(naked)
#endif
ufract rarcmul(ifract y,
	ufract w)
{
#if defined(__POCC__) && envx86
	__asm
	{
		mov	eax,[esp+4]
		mov	edx,eax
		neg	edx
		imul	edx
		shld	edx,eax,2
		shl	eax,2
		add	edx,0x40000000
		push	edx
		push	eax
		call	usqrt64
		add	esp,8
		mul	dword [esp+8]
		mov	eax,edx
		ret
	}
#else
	int64		n;

	(void)muls64s32(n,y,-y);
	int64shl(n,2);
	(void)int64addhl(n,0x40000000,0);
	return ufract32mul(usqrt64(n),w);
#endif
}

/* trigonometry */

/* this LUT is a list of the fraction other than theta in the power series
  x**3/6 - ...
*/
static const uint32	ufsinlut[257] =
{
	0xA55DE731,0xA55DD2CA,0xA55D9596,0xA55D2F95,
	0xA55CA0C7,0xA55BE92C,0xA55B08C5,0xA559FF92,
	0xA558CD95,0xA55772CE,0xA555EF3E,0xA55442E7,
	0xA5526DC8,0xA5506FE5,0xA54E493D,0xA54BF9D3,
	0xA54981A8,0xA546E0BD,0xA5441716,0xA54124B4,
	0xA53E0998,0xA53AC5C5,0xA537593D,0xA533C404,
	0xA530061A,0xA52C1F83,0xA5281042,0xA523D85A,
	0xA51F77CD,0xA51AEE9E,0xA5163CD1,0xA5116269,
	0xA50C5F6A,0xA50733D7,0xA501DFB3,0xA4FC6303,
	0xA4F6BDCA,0xA4F0F00C,0xA4EAF9CD,0xA4E4DB12,
	0xA4DE93DF,0xA4D82438,0xA4D18C22,0xA4CACBA1,
	0xA4C3E2BA,0xA4BCD173,0xA4B597CF,0xA4AE35D5,
	0xA4A6AB88,0xA49EF8F0,0xA4971E10,0xA48F1AEF,
	0xA486EF92,0xA47E9BFF,0xA476203C,0xA46D7C4E,
	0xA464B03C,0xA45BBC0D,0xA4529FC5,0xA4495B6D,
	0xA43FEF09,0xA4365AA2,0xA42C9E3C,0xA422B9E1,
	0xA418AD96,0xA40E7962,0xA4041D4C,0xA3F9995D,
	0xA3EEED9B,0xA3E41A0D,0xA3D91EBC,0xA3CDFBAF,
	0xA3C2B0EE,0xA3B73E81,0xA3ABA470,0xA39FE2C2,
	0xA393F981,0xA387E8B4,0xA37BB065,0xA36F509B,
	0xA362C95F,0xA3561ABB,0xA34944B6,0xA33C475B,
	0xA32F22B1,0xA321D6C3,0xA314639A,0xA306C93E,
	0xA2F907BB,0xA2EB1F18,0xA2DD0F60,0xA2CED89E,
	0xA2C07AD9,0xA2B1F61E,0xA2A34A76,0xA29477EB,
	0xA2857E88,0xA2765E56,0xA2671762,0xA257A9B4,
	0xA2481559,0xA2385A5B,0xA22878C5,0xA21870A2,
	0xA20841FE,0xA1F7ECE3,0xA1E7715D,0xA1D6CF78,
	0xA1C6073F,0xA1B518BE,0xA1A40401,0xA192C913,
	0xA1816801,0xA16FE0D7,0xA15E33A1,0xA14C606B,
	0xA13A6742,0xA1284832,0xA1160348,0xA1039891,
	0xA0F1081A,0xA0DE51EF,0xA0CB761D,0xA0B874B2,
	0xA0A54DBB,0xA0920145,0xA07E8F5E,0xA06AF813,
	0xA0573B71,0xA0435987,0xA02F5263,0xA01B2611,
	0xA006D4A1,0x9FF25E21,0x9FDDC29E,0x9FC90226,
	0x9FB41CC9,0x9F9F1295,0x9F89E399,0x9F748FE2,
	0x9F5F1781,0x9F497A83,0x9F33B8F8,0x9F1DD2EF,
	0x9F07C877,0x9EF199A0,0x9EDB4678,0x9EC4CF0F,
	0x9EAE3375,0x9E9773BA,0x9E808FEC,0x9E69881C,
	0x9E525C5B,0x9E3B0CB7,0x9E239941,0x9E0C020A,
	0x9DF44721,0x9DDC6897,0x9DC4667C,0x9DAC40E2,
	0x9D93F7D8,0x9D7B8B6F,0x9D62FBB9,0x9D4A48C6,
	0x9D3172A7,0x9D18796E,0x9CFF5D2B,0x9CE61DF0,
	0x9CCCBBCE,0x9CB336D7,0x9C998F1C,0x9C7FC4B0,
	0x9C65D7A2,0x9C4BC807,0x9C3195EF,0x9C17416C,
	0x9BFCCA91,0x9BE2316F,0x9BC7761A,0x9BAC98A3,
	0x9B91991C,0x9B767799,0x9B5B342C,0x9B3FCEE7,
	0x9B2447DE,0x9B089F22,0x9AECD4C8,0x9AD0E8E3,
	0x9AB4DB84,0x9A98ACC0,0x9A7C5CAA,0x9A5FEB55,
	0x9A4358D4,0x9A26A53C,0x9A09D09F,0x99ECDB12,
	0x99CFC4A8,0x99B28D75,0x9995358D,0x9977BD03,
	0x995A23ED,0x993C6A5E,0x991E906B,0x99009627,
	0x98E27BA8,0x98C44101,0x98A5E648,0x98876B90,
	0x9868D0EF,0x984A167A,0x982B3C44,0x980C4264,
	0x97ED28EF,0x97CDEFF8,0x97AE9796,0x978F1FDE,
	0x976F88E5,0x974FD2C1,0x972FFD87,0x9710094C,
	0x96EFF626,0x96CFC42B,0x96AF7371,0x968F040E,
	0x966E7616,0x964DC9A2,0x962CFEC5,0x960C1597,
	0x95EB0E2E,0x95C9E8A0,0x95A8A503,0x9587436E,
	0x9565C3F8,0x954426B6,0x95226BC0,0x9500932B,
	0x94DE9D10,0x94BC8984,0x949A589F,0x94780A78,
	0x94559F25,0x943316BE,0x94107159,0x93EDAF0F,
	0x93CACFF6,0x93A7D425,0x9384BBB5,0x936186BC,
	0x933E3552,0x931AC790,0x92F73D8B,0x92D3975D,
	0x92AFD51C,0x928BF6E2,0x9267FCC5,0x9243E6DE,
	0x921FB544
};

ifract fastsin(ufract theta)
{
	ufract		sa, sb;
	unsigned int	t = theta, u;
/*
 theta' = theta{30}

 high two bits:
  00 (0.0pi): theta'' =  theta';    result =  sin(theta'')
  01 (0.5pi): theta'' =  pi-theta'; result =  sin(theta'')
  10 (1.0pi): theta'' =  theta';    result = -sin(theta'')
  11 (1.5pi): theta'' =  pi-theta'; result = -sin(theta'')
*/
	if ((t+=t)&1<<31)
		t =-(ifract)t;
	/* get the two nearest values in the LUT of (theta mod pi/2) and interpolate */
	sb = ufsinlut[u=t>>23], sb-= ufract32mul(sb-ufsinlut[u+1],t<<9);
	/* theta''*(pi/2-theta''**2*...) */
	sa = 0xC90FDAA2-ufract32mul(ufract32mul(t,t)*2,sb);
	/* negative if >= pi */
	return !(theta&1<<31) ? ufract32mul(sa,t):-(ifract)ufract32mul(sa,t);
}

ufract ufractatan2(uint32 x,
	uint32 y)
{
	uint32		t, u, h;

	if (y<=(unsigned)x) /* y>x: return pi/2-atan2(y,x) */
	{
		u = 0;
		if (y==(unsigned)x) /* x=y: 0 ? pi/2:pi/4 -- pol(0,0) is actually undefined */
			{t = 1686629713+(1686629713&-!x); goto endo;}
	} else
		u = y, y = x, x = u, u = 0x80000000u;
	/* r = y/x */
	h = (y=ufract32div(y,x))&-1<<26;
	{ /* r = atan(a)+atan((r-a)/(r*a+1)) */
		static const uint32		hibtable[64] =
		{
			0x00000000,0x03FFEAAB,0x07FF556F,0x0BFDC0C2,
			0x0FFAADDC,0x13F59F0E,0x17EE1826,0x1BE39EBE,
			0x1FD5BA9B,0x23C3F5F6,0x27ADDDD2,0x2B93023C,
			0x2F72F698,0x334D51D3,0x3721AEA5,0x3AEFABBE,
			0x3EB6EBF2,0x4277165F,0x462FD68C,0x49E0DC81,
			0x4D89DCDC,0x512A90DB,0x54C2B665,0x5852100C,
			0x5BD86508,0x5F55812E,0x62C934E5,0x66335515,
			0x6993BB0F,0x6CEA4477,0x7036D325,0x73794D0D,
			0x76B19C16,0x79DFADFC,0x7D03742D,0x801CE39E,
			0x832BF4A7,0x8630A2DB,0x892AECE0,0x8C1AD446,
			0x8F005D5F,0x91DB8F16,0x94AC72CA,0x97731420,
			0x9A2F80E6,0x9CE1C8E7,0x9F89FDC5,0xA22832DC,
			0xA4BC7D19,0xA746F2DE,0xA9C7ABDC,0xAC3EC0FC,
			0xAEAC4C39,0xB110688B,0xB36B31C9,0xB5BCC490,
			0xB8053E2C,0xBA44BC7E,0xBC7B5DEB,0xBEA94145,
			0xC0CE85B9,0xC2EB4ABB,0xC4FFAFFB,0xC70BD54D
		};

		t = hibtable[h>>26];
		y = ufract30div(y-h,((ufract32mul(y,h)+2)>>2)+(1u<<30));
	}
	/* s = sqr(r) */
	h = ufract32mul(y,y);

	/* t+= r**1/1-r**3/3+r**5/5.. */
	t+= y;
	y = ufract32mul(y,h);
	t-= ufract32mul(y,((4294967295.0+3)/3));
	/* the above table is large enough that integer undeflow would occur */
	/* (no further operations needed) */
	t = (t+1)>>1;
endo: /* adjust for ufract format (value is currently in radians) */
	return (!u ? ufract32mul(t,2734261102u)
		:u+-(ifract)ufract32mul(t,2734261102u))+1>>1;
}

ufract ifractatan2(int32 x,
	int32 y)
{
/*
  n,p  2  .5pi   1 p,p
            |
       pi---|--- 0
            |
			I
  n,n  3  1.5pi  4 p,n
*/
	/* y<0: return 2*pi-atan2(x,-y) */
	/* x<0: return pi/2+atan2(y,-x) */
	if (y>=0)
		if (x>=0)
			return ufractatan2(x,y);
		else
			return (1L<<30)+ufractatan2(y,x=-x);
	else if (x>=0)
		return -(ifract)ufractatan2(x,y=-y);
	else
		return 0xC0000000+-(ifract)ufractatan2(y=-y,x=-x);
}

ifract fractasin(ifract x)
{
	uint64		y;
	ufract		a;

	if ((a=ABS(x))>(1<<30))
		return x<0 ? -1<<31:(1u<<31)-1;
	(void)mulu64u32(y,a,a);
	(void)uint64neg(y);
	(void)uint64addhl(y,1u<<60-32,0);
	return ifractatan2(usqrt64(y),x);
}

#if !defined(__POCC__) || !envx86
#undef cntlz32
int cntlz32(uint32 x)
{
	/* divide up logarithm (tree method) */
	if (x&0xFFFF0000u)
	{
		if (x&0xFF000000u)
			if (x&0xF0000000u)
				if (x&0xC0000000u)
					return (long)x>=0;
				else
					return !(x&0x20000000)+2;
			else if (x&0xC000000)
				return !(x&0x8000000)+4;
			else
				return !(x&0x2000000)+6;
		if (x&0xF00000)
			if (x&0xC00000)
				return !(x&0x800000)+8;
			else
				return !(x&0x200000)+10;
		if (x&0xC0000)
			return !(x&0x80000)+12;
		return !(x&0x20000)+14;
	}
	if (x&0xFF00)
		if (x&0xF000)
			if (x&0xC000)
				return !(x&0x8000)+16;
			else
				return !(x&0x2000)+18;
		else if (x&0xC00)
			return !(x&0x800)+20;
		else
			return !(x&0x200)+22;
	if (x&0xF0)
		if (x&0xC0)
			return !(x&0x80)+24;
		else
			return !(x&0x20)+26;
	if (x&12)
		return !(x&8)+28;
	return x&2 ? 30:32-x;
}
#endif

#ifdef TEST_TIME
#include <windows.h>
#include <stdio.h>
/* per million, on a 2.593GHz Pentium 4
  usqrt:	C ms / clocks	Asm ms / cyc	comment
	FPU	-		78-88 / ~227	larger nos. take longer
	IU	54.8 / 142	32.7 / 85	- Asm less than half of usqrt/FPU!
  usqrt64:	C
	FPU	-		100 / 259
	IU	118.7 / 308	64 / 166	- puts the FPU method to shame
*/
int main(int argc,
	char **argv)
{
	int32		x, ltick;
	/* accuracy tests */
	if (usqrt64(-1ull)!=-1u || usqrt64(0xFFFFFFFE00000000ull)!=-1u
			|| usqrt64(0xFFFFFFFD00000003ull)!=-1u
			|| usqrt64(0xFFFFFFFD00000002ull)!=-2u
			|| usqrt64(0xFFFEFFFFFFFFFFFFull)!=0xFFFF8000u
			|| usqrt64(0x3333333333333333ull)!=0x727C9717)
		fprintf(stderr,"## usqrt64 is not producing accurate results!\n");
	printf("waiting 200 ms for any paging to complete...\n\n");
	Sleep(1);
	for(ltick=GetTickCount(); GetTickCount()-ltick<200; )
		;
	for(x=0,ltick=GetTickCount(); x<10000000; ++x)
		usqrt64((int64)x<<32);
	ltick-= GetTickCount();
	printf(
#ifdef NO_FPU
"%d ms computing 10,000,000 usqrt/IU\n",
#else
"%d ms computing 10,000,000 usqrt/FPU\n",
#endif
		-ltick);
	return 0;
}
#elif defined(MAKE_SRHLUT)
#include <math.h>
#include <stdio.h>
#include <fenv.h>
int main(int argc,
	char **argv)
{
	int		x;

	fputs(
"/* square root half (16 bit) table */\n"
"static const uint16	srhlut[256] =\n"
"{\n"
"	1,",
		stdout);
	for(x=1; x<255; ++x)
		printf(x&7 ? "%d,":"\n\t%d,",
			x<64 ? 0:(uint32)(sqrt((x+.5)/256.)*16384+1.5)*4);
	fputs("65535\n};\n",stdout);
	return 0;
}
#endif
