/*
 mfont.c
  (minimal) font loading and rendering

  Created: 2014.02.28 12:43 EST

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <enviro64.h>
#include <stdlib.h>
#include <string.h>

#include "../resex/my_cseries.h"
#include "../resex/resex.h"

/* ---- constants */

/* system font characters */
#define MacSysCharCommand	17
#define MacSysCharCheck		18
#define MacSysCharDiamond	19
#define MacSysCharApple		20

/* a bit of specs from Fonts.r & Fonts.h */
/* these constants & structures summarize, but aren't used in rendering */

/* map flags */
#define famFixedWidth		0x80 /* otherwise, proportional */
#define famIgnoreFractWidths	0x40
#define famIntegralExtra	0x20 /* otherwise, fixed-point */
#define famIgnoreFractEnable	0x10
#define famNoSpacingAdjust	0x8 /* otherwise, adjust char. spacing */
/* reserved flags */
/* table flags */
#define tblHasCharWidths	0x2
#define tblHasImageHeights	0x1

/* ---- typedefs */

typedef struct fontAssc/*(6)*/
{
	int16		size;
	int16		style;
	int16		ID;
} fontAssc;

typedef struct fontFamilyDesc
{
	uint8		famFlags, tblFlags;
	int16		familyID;
	uint16		firstChar, lastChar;
	int16		ascent, descent;
	int16		leading, maxWidth;
	/* offsets */
/*x10*/	int32		widthsTableOffset;
	int32		kerningTableOffset;
	int32		styleTableOffset;
	int16 rsv1;
	int16		exWidBold;
/*x20*/	int16		exWidItalic;
	int16		exWidUnderline;
	int16		exWidOutline;
	int16		exWidShadow;
	int16		exWidCondensed;
	int16		exWidExtended;
	int16 rsv2;
	int32 rsvIntl;
	int16		version;
	/* font association table */
/*x34*/	uint16		asscCount;
	fontAssc	assc[1];
	/* 32-bit offset table if version 2 */
	/* bounding box table: zcount.w style.w, Rect */
} fontFamilyDesc;

#define bmtNoFontExpansion	0x40
#define bmtFixedWidth		0x20 /* otherwise, proportional */
#define bmtColor		0x2 /* otherwise, black on white */
#define bmtSynthetic		0x1

#define bmfHasfctb		0x80 /* has font color table */
#define bmfDepthMask		0xC /* bpp = 1<<x */
#define bmfHasCharWidths	0x2
#define bmfHasImageHeights	0x1

#define bmfValidType(v)		((v&0x9C)!=0x90)
#define bmfValidFlag(v)		!(v&0x70)

typedef struct bmFont
{
	uint8		type; /* required signature: 1xx100xx */
	uint8		flag; /* required signature: x000xxxx */
	uint16		firstChar, lastChar;
	int16		maxWidth, maxKerning;
	int16		negDescent;
	int16		bmWidth, bmHeight;
	uint16		OWTLoc; /* + negDescent as high-word if not -1 */
	int16		ascent, descent, leading;
	int16		wordPitch;
	int16		bm[1];
	/* directly after is the loc. table (lastChar-firstChar+3 ct) */
	/* OWTLoc to offset/width table, then optional char widths/img heights */
} bmFont;

/* ---- code */

typedef struct bmfMemory
{
	uint8		flag, style;
	int16		size;
	uint16		firstChar, lastChar;
	uint16		maxWidth, maxKerning;
	uint16		bmWidth, bmHeight;
	int32		bmSize, pitch;
	int32		wTableOffs; /* in-memory offset */
	int16		ascent, descent, leading;
	uint16		lineHeight; /* calculated */
	int8		*data;
	/* data: location table (uint16*) follows (data+bmSize),
	  width table (int8*) directly follows
	  - 'undefined' char @ (lastChar-firstCharacter+1)*2
	*/
} bmfMemory;

typedef struct ffMemory
{
	uint16		asscCount;
	bmfMemory	font[1];
} ffMemory;

static void b1to8(int8 *d,
	const int8 *s,
	int c)
{
	do
	{
		uint8		v = *s++;

		*d++ = v>>7, v+= v;
		*d++ = v>>7, v+= v;
		*d++ = v>>7, v+= v;
		*d++ = v>>7, v+= v;
		*d++ = v>>7, v+= v;
		*d++ = v>>7, v+= v;
		*d++ = v>>7, v+= v;
		*d++ = v;
	} while(--c);
}

static void b2to8(int8 *d,
	const int8 *s,
	int c)
{
	do
	{
		uint8		v = *s++;

		*d++ = v>>6;
		*d++ = v>>4&3;
		*d++ = v>>2&3;
		*d++ = v&=3;
	} while(--c);
}

static void b4to8(int8 *d,
	const int8 *s,
	int c)
{
	do
	{
		uint8		v = *s++;

		*d++ = v>>4;
		*d++ = v&=15;
	} while(--c);
}

static int loadFont(bmfMemory *f,
	const int8 *_s,
	long z)
{
	const int8	*s;
	uint32		o, v, c;
	uint8		u;
	/* pack flags */
	if ((s=_s)==NULL)
		return -1;
	u = s[1];
	if (!bmfValidType(*s) || !bmfValidFlag(u)
			|| z<=OFFSET(bmFont,bm))
		return 0;
	f->flag = (*s&0x60)|(u&bmfHasfctb)|(u<<2&12);
	f->firstChar = va_be16(s);
	f->lastChar = va_be16(s);
	if (f->lastChar<f->firstChar)
		return 0;
	f->maxWidth = va_be16(s);
	f->maxKerning = va_be16(s);
	v = (int16)va_be16(s);
	f->bmWidth = va_be16(s);
	f->bmHeight = va_be16(s);
	o = va_be16(s);
	if (v!=-1u)
		o|= v<<16;
	if (o>=z)
		goto err;
	f->ascent = va_be16(s);
	f->descent = va_be16(s);
	f->leading = va_be16(s);
	v = va_be16(s)*2;
	u = u>>2&3;
	f->pitch = v*8u>>u;
	f->bmSize = (f->bmHeight*f->pitch+3)&~3;
	c = f->lastChar-f->firstChar+3;
	if (f->bmSize<=0 || (f->data=calloc(1,f->bmSize+c*4))==NULL)
	{
	err:	f->bmHeight = f->bmWidth = 0;
		return f->bmSize>0 ? -2:0;
	}
	v*= f->bmHeight;
	switch(u)
	{
	case 0:	b1to8(f->data,s,v);
		break;
	case 1:	b2to8(f->data,s,v);
		break;
	case 2:	b4to8(f->data,s,v);
		break;
	case 3: memcpy(f->data,s,v);
	}
	f->wTableOffs = f->bmSize+c*2;
	{ //loc table, wid table;
		int16		*d = (int16*)(f->bmSize+f->data);

		do
			*d++ = va_be16(s);
		while(--c);
		memcpy(d,_s+o,(f->lastChar-f->firstChar+3)*2);
	}
	return 1;
}

ffMemory *ffLoad(struct ResFile_Memory *rMap,
	uint8 *name)
{
	mhandle		ffh, h;
	ffMemory	*ff;
	int8		*s;
	uint16		c, i;

	if ((ffh=ResFile_GetResource(rMap,MK4CC('F','O','N','D'),name,0,NULL))==NULL)
		return NULL;
	set_mhandle_lock(ffh,1);
	s = OFFSET(fontFamilyDesc,asscCount)+*ffh;
	if ((c=va_be16(s)+1)==0 || (ff=calloc(sizeof(ffMemory)
			+sizeof(fontAssc)*(c-1),1))==NULL)
		return NULL;
	ff->asscCount = c, i = 0;
	do
	{
		bmfMemory	*f = ff->font+i;
		int		v;

		f->size = va_be16(s);
		f->style = s[1], s+= 2;
		v = va_be16(s);
		if ((h=ResFile_GetResource(rMap,MK4CC('N','F','N','T'),NULL,v,NULL))==NULL
				&& (h=ResFile_GetResource(rMap,MK4CC('F','O','N','T'),
				NULL,v,NULL))==NULL)
			continue;
		set_mhandle_lock(h,1);
		loadFont(f,*h,get_mhandle_size(h));
		dispose_mhandle(h);
	} while(++i<c);
	dispose_mhandle(ffh);
	return ff;
}

/*
inline static int drawGlyph(const grafPort *p,
	const bmfMemory *f)
{
	const int32	*t;
	const int8	*s;
	int32		w, h;

	if (i>f->lastChar || (i-=f->firstChar)<0)
		i = f->lastChar-f->firstChar; //+1 ?
	s = f->data, t = s+f->bmSize;
	w = t[i+1]-t[i], s+= t[i];
	h = f->bmHeight;
	y-= f->ascent, x+= f->maxKerning+f->widthTable[i*2];

	return f->widthTable[i*2+1];
}
*/
