/*
 mgraf.h
  limited graphics library, with some initial basis on QuickDraw

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __MGRAF_H
#pragma once

#include "my_cseries.h"

/* ---- constants */

enum
{ /* pType bits */
	pTypeIndexed,		/* no bits - calls for look-up table */
	pTypeFixed	= 0x1,	/* const table - i.e., grayscale */
	pTypeDirect	= 0x2,	/* no table - gamma ramp applies to device */
	pTypeAlpha	= 0x4,	/* has alpha channel (what's left at top) */
	pTypeLGreen	= 0x8,	/* green channel is 1 bit larger (i.e., 5:6:5) */
	pTypeLSB	= 0x80	/* values in little endian ordering */
};

enum
{
	hvResDefault	= 0x480000 /* 72.0 dpi */
};

#if 0
enum
{ /* bStore values */
	bStorePtrNative,	/* system dependent on storage; baseAddr=ptr */
	bStoreLocked,		/* HLock/LocalLock(etc.) used; baseAddr=ptr */
	bStoreUnlocked		/* HUnlock/LocalUnlock(etc.) used; baseAddr=mhandle */
};
#endif

enum
{ /* pen modes */
	pModeCopy,		/* no transparency (fill) */
	pModeOr,		/* darker pixels merge */
	pModeXor,		/* inversion */
	pModeBic		/* back color (erase) */
};

/* ---- typedefs */

typedef struct rect2dq
{
	ifixed		top, left, bottom, right;
} rect2dq;

/* explicit scanline region (no vertical repeats, smallest size=empty) */
typedef struct SLRgn
{
	int32		size; /* total size, including header */
	rectangle	bBox;
	int16		x[1]; /* lines begin and end with negative space */
} SLRgn;

typedef struct pixMap/*(40+)*/
{ /* aligned like crazy */
	uint16		w, h;
	pt2d16		pOrigin;
	int32		pitch; /* bytes per line */
	uint8		pSize; /* 1, 2, 4, 8, 16 or 32; otherwise, codec type */
	uint8		pType; /* see above */
	uint8		bpp; /* 1-32 */
	uint8		minCmp; /* smallest bit count for a direct component */
/*x10*/	ufixed		hRes, vRes; /* horiz. & vert. ppi resolution */
	int8		*baseAddr; /* access to the pixel data (may be NULL) */
	void		*pHand; /* the applicable mhandle, if used for baseAddr */
	/* start of the applicable tables */
/*x24*/	cTab64		*pTable;
	/* invalidated every time the ct seed changes */
	uint8		*iTable; /* inverse table; minCmp=resolution */
	uint8		*dTable; /* direct table; minCmp=resolution */
} pixMap;

typedef struct grafPort
{
	rectangle	clip;
	pt2d16		origin;
	uint8		pMode;
	uint8		alpha; /* for blending (Copy only) */
	uint32		fgColor, bkColor;
	ifixed		penw, penh;
	pixMap		*map;
} grafPort;

#define DCL_BLT_RTN		__fastcall
#define DEF_BLT_RTN(n)		void DCL_BLT_RTN n(int32 o, int32 c, int8 *a, bltParam *f)
#define TYP_BLT_RTN(n)		void(DCL_BLT_RTN*n)(int32 o, int32 c, int8*, struct bltParam*)

typedef struct bltParam/*(77+)*/
{
	rect32		cl; /* clip - may be narrower than the port clip */
	int32		ox, oy, oz; /* abs origin/size */
	int32		vx, vy, vz; /* routine values */
	uint32		opr, opg, opb;
	uint8		opa;
	/* vy is starting y when rtn is called */
	TYP_BLT_RTN(rtn);

	int8		*p, *a, *e; /* pixels start, current and absolute end */

	const void	*dtb; /* indexed modes only */
	const uint8	*itb; /* indexed modes only */
} bltParam;

typedef TYP_BLT_RTN(bltRoutine);

typedef struct trapezoid/*(72+)*/
{
	int64		x1, dx1; /* manual point-x/slope */
	bltParam	prm;
} trapezoid;

/* ---- macros, inlines */

#define q2dsort2(a,b)		do{ pt2dq _u; \
	if(a.y>b.y) _u=a,a=b,b=_u; \
}while(0)

/* remember to *invert* the fraction to get a full starting line */
#define lineFrac(d,k,f)		(void)(d=(uint16)~(f)|(k)<<16,d-=!++d)

static __forceinline ifixed addu(ifixed a,
	ufixed b)
{
	ifixed		v = a+b;

	return v<a ? IMAX(v):v;
}

static __forceinline ifixed subu(ifixed a,
	ufixed b)
{
	ifixed		v = a-b;

	return v>a ? IMIN(v):v;
}

static __forceinline ufixed idim(ifixed a,
	ifixed b)
{
	return a>b ? a-b:b-a;
}

static __forceinline inline uint32 formDirect(rgb48 c,
	int g,
	int rb)
{
	int		h = 16-rb;

	return (c.red>>h<<g|c.green<<g>>16)<<rb|c.blue>>h;
}

static __forceinline inline uint32 formGray(rgb48 c)
{
	uint32		v, g = c.green;

	v = c.red, v+= (((v+g)>>1)+c.blue)>>1;
	return (v>>1)+g;
}

/* initialization */
inline int rInitClip(const bltParam *f, /* must be valid, locked */
	int32 o,
	int8 **a)
{
	int32		c;

	*a = f->p, o-= c=f->cl.top;
	if (o>0)
		c+= o, o = 0;
	*a+= f->oz*(c-f->oy);
	return -o;
}

inline int rInitClip2(const bltParam *f, /* must be valid, locked */
	int32 o,
	int32 c,
	int8 **a,
	int8 **e)
{
	*e = (*a=f->p)+f->oz*(c-f->oy);
	o-= c=f->cl.top;
	if (o>0)
		c+= o, o = 0;
	*a+= f->oz*(c-f->oy);
	return -o;
}

/* ---- globals */

extern const uint8	Mac8bppRGB12[2*256];
extern const uint8	Mac4bppRGB24[3*16];
extern const uint8	Mac2bppRGB24[3*4];
extern const uint32	pSizeMasks[32];

/* ---- prototypes */

/*
 newPixMap
	Allocates a pixMap, using calloc and the fields given.
	Pass zero for bpp to attempt to match the screen format.
	Call disposePixMap when done.
  Notes:
	This function neither allocates any pixel data nor a table.
	It assumes the pType, based on bpp.
	The initial NULL pTable implies grayscale for indexed modes.
	The pitch is 4-byte aligned.
	The user may modify the structure to see fit.
  Result:
	The map, or NULL on error.
*/
pixMap *newPixMap(int bpp,
	int width,
	int height);
/*
 disposePixMap
	Disposes of a newPixMap-created pixMap.
	Pixels are not disposed if the pHand field is NULL.
*/
int disposePixMap(pixMap*);

/*
 lockPixels
	Locks the pixmap baseAddr, updating the map if necessary.
  Result:
	The pixels base address, or NULL on error.
*/
int8 *lockPixels(pixMap *p);
/*
 unlockPixels
	Unlocks the pixmap baseAddr, updating the map if necessary.
  Result:
	Positive if unlocked, zero if nothing changed, negative on error.
*/
int unlockPixels(pixMap *p);

/*
 openGrafPort
	Initializes a grafPort, with default values and a pixMap.
	Close with closeGrafPort.
  Note:
	It calls newPixMap.
  Result:
	Positive on success; otherwise, an error.
*/
int openGrafPort(grafPort *p,
	int bpp,
	int width,
	int height);
/*
 closeGrafPort
	Disposes of the contents within a grafPort.
*/
int closeGrafPort(grafPort*);

/*
 pixMapToNativeDesc
	Allocates and defines a native platform bitmap/pixmap from a pixMap.
*/
#if defined(WINUSERAPI)
BITMAPINFO *pixMapToNativeDesc(pixMap*);
#elif TARGET_OS_MAC
PixMapHandle pixMapToNativeDesc(pixMap*);
#endif

/* color selection & inverse tables */

/*
 rgbToPenColor/rgb30ToPenColor
	Forms a pixel for the passed pixmap, given a color, filling 32 bits.
	It may build an inverse table for the pixMap.
  Result:
	Can only return zero on error.
*/
uint32 rgbToPenColor(pixMap *m,
	const rgb48*);
uint32 rgb30ToPenColor(pixMap *m,
	rgb30);

/*
 makeITables
	Builds a pixMap's inverse & direct tables, given its pTable.
	minCmp is where the channel resolution resides.
	The res. for a gray table is always 8, though minCmp=0.
  Result:
	1 on success; otherwise, a basic error code.
*/
int makeITables(pixMap *m, /* must be valid */
	int channelRes);

/*
 expandRGB24/expandRGB12
	These expand compacted tables into a cSpec table.
	expandRGB24 takes 3 unaligned bytes each.
	expandRGB12 takes 2 bytes each in the form: 0x0r,0xgb.
*/
void expandRGB24(const uint8 *s,
	cSpec *d,
	int count,
	int startIx);
void expandRGB12(const uint8 *s,
	cSpec *d,
	int count,
	int startIx);

/* graphics */

/* the results for the renderers:
 -2 = memory error (standard)
 -1 = parameter error or data corrupted
  0 = clipped-out or empty
  1 = within bounds (doesn't necessarily mean any pixels changed)
*/

/* regions */

/*
 newRgn
	Creates an empty region; bBox top/bottom are zero.
	Returns NULL on error.
*/
SLRgn *newRgn(void);

/*
 rgnSectLine
	Inserts the integral x-coords of a line into a region.
	Automatically adjusts the size and reallocates into rrgn.
	May automatically create a new region.
	The ptr in rrgn will not become invalid (unless already corrupted).
  Result:
	1 on success; 0 if nothing added; negative on error.
*/
int rgnSectLine(SLRgn **rrgn,
	int16 cliptop,
	int16 clipbot,
	pt2dq p0,
	pt2dq p1);

/*
 rRegion
	Given the current pen mode, paints the area within a region.
  NOTE:
	Framing is not currently implemented, but a hollow pen is recognized.
*/
int rRegion(grafPort *p,
	const SLRgn *rgn);

/*
 rTriangle
	Given the current drawing mode, paints the area within three points.
	This routine ignores the pen width & height.
*/
int rTriangle(grafPort *p,
	pt2dq p0,
	pt2dq p1,
	pt2dq p2);

/*
 rPolygon
	Given the current pen mode, paints a polygon.
	If the penw is valid, a line frame is drawn.
	(A "single-pixel" frame is drawn if only the penh is invalid.)
*/
int rPolygon(grafPort *p,
	const pt2dq *ep,
	int count);
/*
 rLine
	Like QuickDraw, paints a rectangular pen between two pens.
	If the pen is invalid, a "single-pixel" line is drawn.
*/
int rLine(grafPort *p,
	pt2dq p0,
	pt2dq p1);
/*
 rOval
	Paints the area within an oval, given rectangular bounds.
	If the pen is valid, only a frame is drawn, midway inset/outset.
*/
int rOval(grafPort *p,
	const rect2dq *ov);
/*
 rRect
	Paints the area within a rectangle.
	If the pen is valid, only a frame is drawn, midway inset/outset.
*/
int rRect(grafPort *p,
	const rect2dq *rc);
/*
 rRoundRect
	Paints the area within a rectangle with round edges (rx/ry).
	If the pen is valid, only a frame is drawn, midway inset/outset.
*/
int rRoundRect(grafPort *p,
	const rect2dq *rc,
	ifixed rx,
	ifixed ry);

/* low-level routines
 (only provided publicly for performance)
*/

/*
 rTrapezoid
	Renders a trapezoid.
	The prm field MUST BE initialized with rInitBlt.
*/
void rTrapezoid(trapezoid*, /* must be valid */
	pt2dq p0,
	pt2dq p1);
/*
 rInitBlt
	Initializes a bltParam, with vertical clipping.
	It calls rGetBltRtn, and lockPixels as necessary.
  NOTE:
	This now takes integral y0/y1, not fixed point values.
  Result:
	The clipping y-skip, or -1 on error or hollow.
*/
/* returns clipping y-skip */
int rInitBlt(bltParam *f, /* must be valid */
	grafPort *p, /* must be valid */
	int y0,
	int y1);

#define __MGRAF_H
#endif
