/*
 spanner.c
  portable graphics and input demo

  Created: 2014.02.15 20:58 EST
  Revised: 2014.03.11

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
 Pass an integer to set the image bit depth.

 The following keys set coordinates: A, S, D, F, Z, X, C, V.
 The following keys set the line color: W, Y, R, N, G, B, K, O.
 Hold down L to see the triangle pair at other chromatic shades via mouse.
 Press space to toggle animation.
 Press I to toggle inversion.
 Press ESC to turn off the tick timer.

 spanner was originally written in M68K Asm for Mac OS.
*/

#include <enviro64.h>
#include <stdio.h>

#include <windows.h>
#include <process.h>

#include "my_cseries.h"
#include "mgraf.h"
#include "mevent.h"

#define devWindowName "spanner"

#define DEFAULT_BPP	24

/* ---- required suite */

void global_idle_proc(void)
{
}

void stop_interface_fade(void)
{
}

uint8 *getpstr(void *d,
	short ID,
	int x)
{
	if (!d)
		return NULL;
	*(int8*)d = 0;
	return d;
}

static uint8	_dbgTemp[1023], debugSet;

#define ATTASKLEVEL0()		1
#define sys_debugger_available	0

int dprintf(const char *form,
	...)
{
	va_list		args;
	char		*s;
	int		len;

	va_start(args,form);

	if ((len=vsprintf(1+(char*)_dbgTemp,form,args))>255)
		len = 255;

	va_end(args);

	s = memchr(1+_dbgTemp,';',*_dbgTemp=len);
	//_dbgTemp[len+1] = 0; -- already set by vsprintf

	/* if first ';' follows "g;" */
	if (!ATTASKLEVEL0() || !s || s[1]!='g') // || s[2]!=';')
	{
		if (sys_debugger_available)
		{
#if TARGET_OS_MAC
			DebugStr(_dbgTemp);
#endif
		} else
			debugSet = 1;
	}
	if (s)
		*s = 0;
	if (ATTASKLEVEL0())
		printf("## %s\n",
			1+_dbgTemp);
	return len;
}

static int printOSMessage(int32 err,
	const char *caller,
	const char *defaultMsg)
{
	LPVOID		b = NULL;
	int		r;

	if (err || !defaultMsg)
		FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER|FORMAT_MESSAGE_FROM_SYSTEM,
			NULL,err,
			MAKELANGID(LANG_NEUTRAL,SUBLANG_DEFAULT),
			(LPTSTR)&b,0,NULL);
	r = dprintf("%s: %s",
		caller,b ? b:defaultMsg);
	if (b)
		LocalFree(b);
	return r;
}

/* ---- code */

#define screen_window world_screen_window
HWND		screen_window;

static WNDCLASS		wc;
static HPALETTE		screen_pal;
static HBITMAP		screen_dib;
static grafPort		screen_world;
static int		screen_x, screen_y;
static int8		screen_fs;

void SetEntries(short o,
	short count,
	const PALETTEENTRY *cs)
{
	RGBQUAD		b[256], *d;
	HDC		hdc, mdc;
	short		c;

	if (!cs || (c=count)<=0)
		return;
	hdc = GetDC(screen_window);
	if ((mdc=CreateCompatibleDC(hdc))==NULL)
		return;
	SelectObject(mdc,screen_dib);
	d = b;
	do
	{
		d->rgbRed = cs->peRed;
		d->rgbGreen = cs->peGreen;
		d->rgbBlue = cs->peBlue;
		d->rgbReserved = 0;
	} while(++d,++cs,--c);
	SetDIBColorTable(mdc,o,count,b);
	DeleteDC(mdc);
#ifndef _WIN32_WCE
	if (screen_pal)
		UnrealizeObject(screen_pal);
#endif
	if (!screen_pal)
	{
		LOGPALETTE	*tmppal;

		if ((tmppal=calloc(sizeof(LOGPALETTE),1))!=NULL)
		{
			tmppal->palVersion = 0x300;
			tmppal->palNumEntries = 256;
			screen_pal = CreatePalette(tmppal);
			free(tmppal);
		}
		if (!screen_pal)
			return;
	}
	SetPaletteEntries(screen_pal,o,count,cs);
	SelectPalette(hdc,screen_pal,FALSE);
	RealizePalette(hdc);
	ReleaseDC(screen_window,hdc);
}

uint32		mTickCount;
uint16		mEventState = btnState;
uint8		mKeyMap[20];
static int8	altDown;
static pt2d16	mMouseLoc;

static const char	*base36 = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";

static char *hexToStr(char *d,
	const uint8 *s,
	int z)
{
	if (!d || z<=0)
		return d;
	do
		*d++ = base36[*s>>4], *d++ = base36[*s++&15];
	while(--z);
	return d;
}

static void resizeWindow(int w,
	int h)
{
	RECT		bounds;
	UINT		swp_flags = SWP_NOCOPYBITS|SWP_SHOWWINDOW;
	int		x, y;

	bounds.left = screen_x;
	bounds.top = screen_y;
	bounds.right = screen_x+w;
	bounds.bottom = screen_y+h;

	AdjustWindowRect(&bounds,GetWindowLong(screen_window,GWL_STYLE),0);

	if (screen_fs)
	{
		x = (GetSystemMetrics(SM_CXSCREEN)-w)/2;
		y = (GetSystemMetrics(SM_CYSCREEN)-h)/2;
	} else if (screen_x || screen_y)
		x = bounds.left, y = bounds.top;
	else
		x = y =-1, swp_flags|= SWP_NOMOVE;
	SetWindowPos(screen_window,screen_fs ? HWND_TOPMOST:HWND_NOTOPMOST,
		x,y,w,h,swp_flags);
	SetForegroundWindow(screen_window);
}

/* ---- active window */

static void MacOSVBL(timerEntry *this)
{ /* the Mac OS tick rate is 16625 mcs, or 133/8 ms */
	++mTickCount;
	(void)int64addu32(this->mcsremaining,16625);
	if (int64n(this->mcsremaining))
		int64xset(this->mcsremaining,0);
}

static uint8	timerCount = 1; /* NOTE: 0 counts as 256! */
#define MAXTIMERS	32
static timerEntry	timers[MAXTIMERS] = {_int64xval(0), MacOSVBL};

int PrimeTimer(timerEntry *this,
	int64 mcs)
{
	int64		n;

	if (!this)
		return -1;
	/* quasi-locked read */
	*(double*)&n = *(double*)&this->mcsremaining;
	if (!int64n(n))
		return 0; /* already primed */
	(void)int64add(n,mcs);
	if (int64n(n))
		int64xset(n,0);
	/* quasi-locked write */
	*(double*)&this->mcsremaining = *(double*)&n;
	return 1;
}

timerEntry *InstallTimer(timerRoutine rtn)
{
	timerEntry	*t;

	if (!rtn)
		return NULL;
	if (timerCount>=MAXTIMERS)
		return NULL; /* maxed out */
	t = timers+timerCount;
	int64xset(t->mcsremaining,-1);
	t->routine = rtn;
	++timerCount;
	return t;
}

int RemoveTimer(timerEntry *this)
{
	timerEntry	*t;
	uint8		c;

	if (!this)
		return -1;
	t = timers, c = timerCount;
	if (c<2)
		return 0;
	if (c>MAXTIMERS)
		timerCount = c = MAXTIMERS;
	do
		if (t==this)
		{
			timerCount = --c, ++t;
			memcpy(t-1,t,c*sizeof(*t));
			//t+= c;
			//int64xset(t->mcsremaining,-1);
			//t->routine = NULL;
			return 1;
		}
	while(++t,--c);
	return 0;
}

static int8	timer_active = TRUE;
static HANDLE	timer_thread;
static int32	timer_start;
static int8	mQuit;

#define DISPLAY_STATS		1

static void updateRect(int x,
	int y,
	int w,
	int h)
{
	HDC		hdc = GetDC(screen_window);
	HDC		mdc = CreateCompatibleDC(hdc);

	SelectObject(mdc,screen_dib);
	BitBlt(hdc,x,y,w,h,mdc,x,y,SRCCOPY);
	DeleteDC(mdc);
#if DISPLAY_STATS
	{
		char		b[sizeof(mKeyMap)*2+64];
		int32		v = TickCount();

		TextOut(hdc,x+6,y+h,b,sizeof(mKeyMap)*2
			+sprintf(hexToStr(b,mKeyMap,sizeof(mKeyMap))," %.2fs - %.3f ms ",
			v/60.,((int32)GetTickCount()-timer_start)/(double)v));
	}
#endif
	ReleaseDC(screen_window,hdc);
}

static void update_any_window(HWND window)
{
	PAINTSTRUCT	ps;

	BeginPaint(window,&ps);
	EndPaint(window,&ps);
}

#define MODKEYS_6L	(0x80>>(55&7) \
	|(0x80>>(56&7)|0x80>>(57&7)|0x80>>(59&7) \
	|0x80>>(60&7)|0x80>>(61&7)|0x80>>(62&7))<<8 \
	|(0x80>>(66&7))<<16)

uint8 *MGetKeys(uint8 outmap[20])
{
	uint8		vkmap[256];

	if (!outmap)
		return NULL;
	/* only GetKeyboardState will give us independent left/right keys */
	GetKeyboardState(vkmap);
	*(int32*)(mKeyMap+6)&=~MODKEYS_6L;
	mKeyMap[6]|= vkmap[VK_LMENU]>>7&1;
	mKeyMap[7]|= (vkmap[VK_LSHIFT]&128)|(vkmap[VK_RSHIFT]>>4&8)
		|(vkmap[VK_LCONTROL]>>3&16)|(vkmap[VK_RCONTROL]>>6&2)
		|(vkmap[VK_RMENU]>>5&4)|(GetKeyState(VK_CAPITAL)<<6&64);
	memcpy(mKeyMap,outmap,20);
	return outmap;
}

int OSPollKeyMouse(boolean down)
{
#if TARGET_OS_MAC
	EventRecord	event;	

	event.what = 0;
	if (down)
		return GetNextEvent(mDownMask|keyDownMask|autoKeyMask,&event)
			&& event.what;
	else
		return GetNextEvent(mDownMask|mUpMask|keyDownMask|keyUpMask|autoKeyMask,&event)
			&& event.what && event.what!=mouseUp && event.what!=keyUp;
#else
	uint8		*s = mKeyMap;
	int		r = *(int16*)s;

	r|= ((int16*)s)[1];
	r|= ((int16*)s)[2];
	r|= s[6]&~3; /* [7] are all modifiers */
	r|= s[8], r|= s[9]&31; /* ignoring volume buttons */
	r|= ((int16*)s)[5];
	r|= ((int16*)s)[6];
	r|= ((int16*)s)[7];
	r|= ((int16*)s)[8]; /* the rest are interface */
	return down ? r:!r;
#endif
}

uint16 updateEventState(void)
{
	uint16		state = 0;

	*(int32*)(mKeyMap+6)&=~MODKEYS_6L;
	if (GetKeyState(VK_SHIFT)>>1)
	{
		state = shiftKey;
		if (GetKeyState(VK_LSHIFT)>>1)
			mKeyMap[7]|= 0x80>>(56&7);
		if (GetKeyState(VK_RSHIFT)>>1)
			mKeyMap[7]|= 0x80>>(60&7);
	}
	if (GetKeyState(VK_CAPITAL)&1) /* sticky capsLock */
		state|= capsLock, mKeyMap[7]|= 0x80>>(57&7);
	if (GetKeyState(VK_CONTROL)>>1)
	{
		state|= controlKey;
		if (GetKeyState(VK_LCONTROL)>>1)
			mKeyMap[7]|= 0x80>>(59&7);
		if (GetKeyState(VK_RCONTROL)>>1)
			mKeyMap[7]|= 0x80>>(62&7);
	}
	if (altDown)
		state|= cmdOrAltKey;
	if (altDown || GetKeyState(VK_MENU)>>1)
	{
		if (GetKeyState(VK_LMENU)>>1)
			mKeyMap[6]|= 0x80>>(55&7);
		if (GetKeyState(VK_RMENU)>>1)
			mKeyMap[7]|= 0x80>>(61&7);
	}
	mEventState&= btnState;
	return mEventState|= state;
}

static LRESULT CALLBACK WndProc(HWND window,
	UINT what,
	WPARAM wParam,
	LPARAM lParam)
{
	switch(what)
	{
	case WM_CREATE:
		//SetTimer(window,idMCSTimer,1,NULL);
		break;
	case WM_DESTROY:
		PostQuitMessage(0);
		break;
	case WM_CLOSE:
		//KillTimer(window,idMCSTimer);
		DestroyWindow(window);
		break;
	//case WM_PALETTECHANGED: -- system colors changed
	//	break;
	case WM_PAINT:
		update_any_window(window);
		break;
	case WM_MOUSEMOVE:
		mMouseLoc.x = (int16)LOWORD(lParam);
		mMouseLoc.y = (int16)HIWORD(lParam);
		break;
	//case WM_CHAR:
	//case WM_DEADCHAR: -- dead keys are accent modifiers (e.g., umlaut)
	//case WM_SYSDEADCHAR:
	//case WM_ERASEBKGND: /* Basilisk II: to prevent flashing */
	//	return 1;
	case WM_LBUTTONDOWN:
	case WM_LBUTTONUP:
	case WM_LBUTTONDBLCLK:
		mEventState = (uint16)((mEventState&~btnState)
			|((~what)<<7&btnState));
		break;
	case WM_SYSKEYDOWN:
	case WM_KEYDOWN:
	case WM_SYSKEYUP:
	case WM_KEYUP:
	{ /* need to trap the ALT and WIN keys! .. no success on WIN keys */
		uint16		state;
		uint8		mvk, down = (uint8)(1&~what);

		if (down && lParam&0x40000000)
			return 0; /* auto-repeat */
		if (wParam>226u || (mvk=mkvtab_win[wParam])>=MKC)
			break;
		if (!(lParam&0x1000000))
			switch(wParam)
			{ /* numlock-off numpad translation */
			case VK_CLEAR: mvk = 87; break; /* KP5 */
			case VK_PRIOR: mvk = 92; break; /* KP9 */
			case VK_NEXT: mvk = 85; break; /* KP3 */
			case VK_END: mvk = 83; break; /* KP1 */
			case VK_HOME: mvk = 89; break; /* KP7 */
			case VK_LEFT: mvk = 86; break; /* KP4 */
			case VK_RIGHT: mvk = 88; break; /* KP6 */
			case VK_UP: mvk = 91; break; /* KP8 */
			case VK_DOWN: mvk = 84; break; /* KP2 */
			case VK_INSERT: mvk = 82; break; /* KP0 */
			case VK_DELETE: mvk = 65;
			}
		else if (wParam==VK_RETURN)
			mvk = 76; /* enter */
		altDown = (what&~1)==WM_SYSKEYDOWN && lParam&0x20000000;
		if (down)
			mKeyMap[mvk>>3]|= 0x80>>(mvk&7);
		else
			mKeyMap[mvk>>3]&=~(0x80>>(mvk&7));
		state = updateEventState();
		/* as an EventRecord post: state&=~eventModifiedSwitch(mvk); */
		/* cmd/alt-Q */
		if (altDown && mvk==12)
			PostQuitMessage(0);
		else if (mvk==53)
			timer_active = FALSE;
	case WM_HOTKEY:
	//case WM_KILLFOCUS:
	//case WM_SETFOCUS:
	//case WM_CONTEXTMENU: /* if you want to disable contextual menu on titlebar */
		break;
	}
	case WM_NCACTIVATE:
	case WM_ACTIVATE:
		memset(mKeyMap,0,sizeof(mKeyMap));
	case WM_SYSCOMMAND:
		if ((wParam&0xFFF0)==SC_KEYMENU) /* alt-space et al. */
			break;
	default:
/*	case WM_ACTIVATEAPP:
	case WM_WINDOWPOSCHANGING:
	case WM_WINDOWPOSCHANGED:
	case WM_MOVE:
	case WM_SIZE:
	case WM_SETFOCUS:
	case WM_KILLFOCUS:
	case WM_ERASEBKGND:
	case WM_SHOWWINDOW:
	case WM_SETCURSOR:
	case WM_GETICON:
	case WM_NCCREATE:
	case WM_NCCALCSIZE:
	case WM_NCHITTEST:
	case WM_NCPAINT:
	case WM_NCMOUSEMOVE:
	case WM_NCLBUTTONDOWN:
	case WM_SYNCPAINT:
	case WM_IME_SETCONTEXT:
	case WM_IME_NOTIFY:
	case WM_NCMOUSELEAVE:
	case WM_CAPTURECHANGED:
	case WM_MOUSEACTIVATE:
	case WM_NULL:
*/		return DefWindowProc(window,what,wParam,lParam);
	}
	return 0;
}

static int setVideoMode(int depth,
	int width,
	int height)
{
	HDC		hdc;
	BITMAPINFO	*binfo;
	int32		err = 0;

	if (screen_dib)
		{DeleteObject(screen_dib); screen_dib = NULL;}
	if (depth<0)
	{
		closeGrafPort(&screen_world);
		if (screen_window)
			{DestroyWindow(screen_window); screen_window = NULL;}
		if (screen_pal)
			{DeleteObject(screen_pal); screen_pal = NULL;}
		return 1;
	}
	if (screen_window)
		resizeWindow(width,height);
	else
	{ /* create device window */
		screen_window = CreateWindow("myDevWindow",
			devWindowName, 
			WS_CAPTION|WS_POPUPWINDOW|WS_MINIMIZEBOX|WS_VISIBLE,
			0,0,width,height,
			HWND_DESKTOP,NULL,wc.hInstance,NULL);
		if (!screen_window)
			return 0;
	}
	openGrafPort(&screen_world,((depth-1)&31)+1,width,height);

	if (screen_world.map && !screen_world.map->pTable)
		if (depth==8)
		{
			if ((screen_world.map->pTable=calloc(8,257))==NULL)
				err = ERROR_NOT_ENOUGH_MEMORY;
			else
			{
				expandRGB12(Mac8bppRGB12,screen_world.map->pTable->colors,256,0);
				screen_world.map->pTable->count = 256;
			}
		} else if (depth==4)
			if ((screen_world.map->pTable=calloc(8,17))==NULL)
				err = ERROR_NOT_ENOUGH_MEMORY;
			else
			{
				expandRGB24(Mac4bppRGB24,screen_world.map->pTable->colors,16,0);
				screen_world.map->pTable->count = 16;
			}

	if (!err && (binfo=pixMapToNativeDesc(screen_world.map))==NULL)
		err = ERROR_NOT_ENOUGH_MEMORY;
	if (!err)
	{
		hdc = GetDC(screen_window);
		screen_dib = CreateDIBSection(hdc,binfo,DIB_RGB_COLORS,
			(void**)&screen_world.map->baseAddr,NULL,0);
		err = GetLastError();
		free(binfo);

		ReleaseDC(screen_window,hdc);
	}
	if (!screen_dib)
	{
		printOSMessage(err,"setVideoMode","Depth not supported for DIBs");
		return 0;
	}
	return 1;
}

#if 0
static void setGrayPalette(void)
{
	PALETTEENTRY	b[256], *a;
	uint8		i;

	a = b, i = 0;
	do
		a->peRed = a->peGreen = a->peBlue = i, a->peFlags = PC_NOCOLLAPSE;
	while(++a,++i);
	SetEntries(0,256,b);
}
#endif

#define SCRW			640
#define SCRH			480

static void registerWindowClass(HINSTANCE devInstance)
{
	wc.style = CS_OWNDC;
	wc.lpfnWndProc = WndProc;
	wc.cbWndExtra = wc.cbClsExtra = 0;
	wc.hInstance = devInstance;
	wc.hIcon = LoadIcon(NULL,IDI_APPLICATION);
	wc.hCursor = LoadCursor(devInstance,MAKEINTRESOURCE(8001));//NULL,IDC_ARROW);
	wc.hbrBackground = (HBRUSH)GetStockObject(BLACK_BRUSH);
	wc.lpszMenuName = NULL;
	wc.lpszClassName = "myDevWindow";
	RegisterClass(&wc);
}

static void event_loop(void *param)
{
	int32		ltick = GetTickCount()+1;
(void)param;
	while(timer_active)
	{
		int32		ticks = GetTickCount(), mcsd = (ticks-ltick)*1000u;

		ltick = ticks;
		if (mcsd>0)
		{
			timerEntry	*t = timers;
			uint8		c = timerCount;

			do
				if (!int64n(t->mcsremaining))
				{
					(void)int64subu32(t->mcsremaining,mcsd);
					if (int64n(t->mcsremaining))
						t->routine(t);
				}
			while(++t,--c);
		}
		SleepEx(1,0);
	}
	timer_thread = NULL;
	_endthread();
}

int WINAPI WinMain(HINSTANCE hInstance,
	HINSTANCE hPrevInstance,
	LPSTR lpCmdLine,
	int iCmdShow)
{
	rect2dq		ov;
	pt2dq		tr[4], ln[2];
	uint32		ltick = 0, lineColor;
	uint32		depth = !lpCmdLine ? 0:atoi(lpCmdLine);
	int8		animate = 1, invert = 0, hasDrawn = 0, sleepdec = 0;
(void)hPrevInstance; (void)iCmdShow;
	registerWindowClass(hInstance);

	freopen("stdout.txt","w",stdout);
	//freopen("stderr.txt","w",stderr);
	if (setVideoMode(depth ? depth:DEFAULT_BPP,SCRW,SCRH)<=0)
	{
		dprintf("# error setting %dx%d %dbpp mode",
			SCRW,SCRH,depth);
		goto end;
	}
	//setGrayPalette();
	timer_thread = (HANDLE)_beginthread(event_loop,0,NULL);
	timer_start = GetTickCount();
	/* custom clipping works now */
	screen_world.clip.top = 8;

	/* initialize the graphics */
	memset(lockPixels(screen_world.map),
		rgb30ToPenColor(screen_world.map,0),
		screen_world.map->pitch*screen_world.map->h);
	tr->x = 0x08A8000;
	tr->y = 0x0008000;
	tr[1].x = 0x1248000;
	tr[1].y = 0x01B8000;
	tr[2].x = 0x1208000;
	tr[2].y = 0x0C08000;
	tr[3].x = 0x0718000;// rand()%(SCRW*0x4000)+SCRW*0x8000;
	tr[3].y = 0x0C88000;// rand()%(SCRH*0x4000)+SCRH*0x8000;
	ln->x = 0x1BE8000;
	ln->y = 0x1248000;
	ln[1].x = 0x24B8000;
	ln[1].y = 0x18A8000;
	ov.left = 0x10E8000;
	ov.top = 0x1448000;
	ov.right = 0x19B8000;
	ov.bottom = 0x1AA8000;
	/* initially white */
	lineColor = rgb30ToPenColor(screen_world.map,~0);

	while(!mQuit)
	{
		int32		tick;
		MSG		msg;

		while(PeekMessage(&msg,NULL,0,0,PM_REMOVE)) 
		{
			if (msg.message==WM_QUIT) 
				mQuit = TRUE;
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
		tick = TickCount(); //GetTickCount()>>4;
		if (tick==ltick)
		{
			Sleep(sleepdec^=1);
			continue;
		}
		ltick = tick;
		/* update for both left and right modifier keys */
		MGetKeys(mKeyMap);

		if (hasDrawn)
		{
			screen_world.alpha = 255;
			screen_world.pMode = pModeCopy;
			screen_world.fgColor = rgb30ToPenColor(screen_world.map,0);
			rPolygon(&screen_world,tr,4);
			rLine(&screen_world,*ln,ln[1]);
			rTriangle(&screen_world,tr[2],tr[3],*tr);
			rRoundRect(&screen_world,&ov,16<<16,16<<16);
		}
		if (animate&1)
		{
			tr->x+= (int16)rand();
			tr->y+= (int16)rand();
			tr[1].x+= (int16)rand();
			tr[1].y+= (int16)rand();
			tr[2].x+= (int16)rand();
			tr[2].y+= (int16)rand();
			ln->x+= (int16)rand();
			ln->y+= (int16)rand();
			ln[1].x+= (int16)rand();
			ln[1].y+= (int16)rand();
			ov.left+= (int16)rand();
			ov.top+= (int16)rand();
			ov.right+= (int16)rand();
			ov.bottom+= (int16)rand();
		}
		invert&=~32, invert|= _MKDOWN(mKeyMap,MKV_I);
		if ((invert&96)==32)
			invert^= 1;
		invert&=~64, invert|= invert<<1&64;
		screen_world.pMode = invert&1 ? pModeXor:pModeCopy;
		/* space, -/bksp, = */
		animate&=~84, animate|= (mKeyMap[6]&80)|_MKDOWN(mKeyMap,MKV_HYPHEN_USC)
			|(mKeyMap[3]>>5&4);
		if ((animate&192)==64)
			animate^= 1;
		if ((animate&48)==16) /* - key to decrease pensize */
			screen_world.penw = screen_world.penh-= screen_world.penh<0 ? 0:0x4000;
		if ((animate&12)==4) /* = key to increase pensize */
			screen_world.penw = screen_world.penh+= 0x4000;
		animate&=~168, animate|= animate<<1&168;
		screen_world.fgColor = 0x3FF;
		if (_MKDOWN(mKeyMap,MKV_L)) /* L to set the intensity (by v) */
			screen_world.fgColor = mMouseLoc.y*2;
		screen_world.fgColor = rgb30ToPenColor(screen_world.map,
			screen_world.fgColor*_PIXEL30(1,0,1)
			+MIN(mMouseLoc.x,255)*_PIXEL30(0,4,0)+_PIXEL30(0,3,0));
		/* A, S, D, and F for the quadrilateral */
		if (_MKDOWN(mKeyMap,MKV_A))
			tr[0].x = mMouseLoc.x<<16|0x8000, tr[0].y = mMouseLoc.y<<16|0x8000;
		if (_MKDOWN(mKeyMap,MKV_S))
			tr[1].x = mMouseLoc.x<<16|0x8000, tr[1].y = mMouseLoc.y<<16|0x8000;
		if (_MKDOWN(mKeyMap,MKV_D))
			tr[2].x = mMouseLoc.x<<16|0x8000, tr[2].y = mMouseLoc.y<<16|0x8000;
		if (_MKDOWN(mKeyMap,MKV_F))
			tr[3].x = mMouseLoc.x<<16|0x8000, tr[3].y = mMouseLoc.y<<16|0x8000;
		/* Z and X for the line coords */
		if (_MKDOWN(mKeyMap,MKV_Z))
			ln->x = mMouseLoc.x<<16|0x8000, ln->y = mMouseLoc.y<<16|0x8000;
		if (_MKDOWN(mKeyMap,MKV_X))
			ln[1].x = mMouseLoc.x<<16|0x8000, ln[1].y = mMouseLoc.y<<16|0x8000;
		/* C and V for the oval coords */
		if (_MKDOWN(mKeyMap,MKV_C))
			ov.left = mMouseLoc.x<<16|0x8000, ov.top = mMouseLoc.y<<16|0x8000;
		if (_MKDOWN(mKeyMap,MKV_V))
			ov.right = mMouseLoc.x<<16|0x8000, ov.bottom = mMouseLoc.y<<16|0x8000;
		if (_MKDOWN(mKeyMap,MKV_W))
			lineColor = rgb30ToPenColor(screen_world.map,~0);
		if (_MKDOWN(mKeyMap,MKV_Y))
			lineColor = rgb30ToPenColor(screen_world.map,_PIXEL30(1023,1023,0));
		if (_MKDOWN(mKeyMap,MKV_M))
			lineColor = rgb30ToPenColor(screen_world.map,_PIXEL30(1023,0,1023));
		if (_MKDOWN(mKeyMap,MKV_R))
			lineColor = rgb30ToPenColor(screen_world.map,_PIXEL30(1023,0,0));
		if (_MKDOWN(mKeyMap,MKV_N))
			lineColor = rgb30ToPenColor(screen_world.map,_PIXEL30(0,1023,1023));
		if (_MKDOWN(mKeyMap,MKV_G))
			lineColor = rgb30ToPenColor(screen_world.map,_PIXEL30(0,1023,0));
		if (_MKDOWN(mKeyMap,MKV_B))
			lineColor = rgb30ToPenColor(screen_world.map,_PIXEL30(0,0,1023));
		if (_MKDOWN(mKeyMap,MKV_K))
			lineColor = rgb30ToPenColor(screen_world.map,0);
		if (_MKDOWN(mKeyMap,MKV_O))
			lineColor = rgb30ToPenColor(screen_world.map,_PIXEL30(1023,511,0));
		rPolygon(&screen_world,tr,4);
		screen_world.fgColor = lineColor;
		rLine(&screen_world,*ln,ln[1]);
		screen_world.alpha = 255/3;
		screen_world.fgColor = rgb30ToPenColor(screen_world.map,
			_PIXEL30(1023*.75,1023*.75,1023*.75));
		rRoundRect(&screen_world,&ov,16<<16,16<<16);
		rTriangle(&screen_world,tr[2],tr[3],*tr);

		hasDrawn = 1;

		updateRect(0,0,screen_world.map->w,
			screen_world.map->h-40);
		Sleep(5);
	}
end:	setVideoMode(-1,0,0);
	timer_active = FALSE;
	//while(timer_thread)
	//	Sleep(50);
	//Sleep(10);
	if (!ftell(stdout))
		fclose(stdout), remove("stdout.txt");
	return 0;
}
